@@ -17,6 +17,9 @@
 -- | Generation of random shrinkable, showable functions.
 -- See the paper \"Shrinking and showing functions\" by Koen Claessen.
 --
+-- __Note__: most of the contents of this module are re-exported by
+-- "Test.QuickCheck". You probably do not need to import it directly.
+--
 -- Example of use:
 --
 -- >>> :{
@@ -59,6 +62,7 @@ module Test.QuickCheck.Function
 import Test.QuickCheck.Arbitrary
 import Test.QuickCheck.Poly
 
+import Control.Applicative
 import Data.Char
 import Data.Word
 import Data.List( intersperse )
@@ -71,6 +75,8 @@ import qualified Data.Sequence as Sequence
 import Data.Int
 import Data.Complex
 import Data.Foldable(toList)
+import Data.Functor.Identity
+import qualified Data.Monoid as Monoid
 
 #ifndef NO_FIXED
 import Data.Fixed
@@ -180,6 +186,12 @@ functionMapWith function g h f = Map g h (function (\b -> f (h b)))
 instance Function () where
   function f = Unit (f ())
 
+instance Function a => Function (Const a b) where
+  function = functionMap getConst Const
+
+instance Function a => Function (Identity a) where
+  function = functionMap runIdentity Identity
+
 instance (Function a, Function b) => Function (a,b) where
   function = functionPairWith function function
 
@@ -337,6 +349,34 @@ instance Function Word32 where
 instance Function Word64 where
   function = functionIntegral
 
+-- instances for Data.Monoid newtypes
+
+instance Function a => Function (Monoid.Dual a) where
+  function = functionMap Monoid.getDual Monoid.Dual
+
+instance Function Monoid.All where
+  function = functionMap Monoid.getAll Monoid.All
+
+instance Function Monoid.Any where
+  function = functionMap Monoid.getAny Monoid.Any
+
+instance Function a => Function (Monoid.Sum a) where
+  function = functionMap Monoid.getSum Monoid.Sum
+
+instance Function a => Function (Monoid.Product a) where
+  function = functionMap Monoid.getProduct Monoid.Product
+
+instance Function a => Function (Monoid.First a) where
+  function = functionMap Monoid.getFirst Monoid.First
+
+instance Function a => Function (Monoid.Last a) where
+  function = functionMap Monoid.getLast Monoid.Last
+
+#if MIN_VERSION_base(4,8,0)
+instance Function (f a) => Function (Monoid.Alt f a) where
+  function = functionMap Monoid.getAlt Monoid.Alt
+#endif
+
 -- poly instances
 
 instance Function A where
@@ -454,6 +494,9 @@ shrinkFun shr (Map g h p) =
 data Fun a b = Fun (a :-> b, b, Shrunk) (a -> b)
 data Shrunk = Shrunk | NotShrunk deriving Eq
 
+instance Functor (Fun a) where
+  fmap f (Fun (p, d, s) g) = Fun (fmap f p, f d, s) (f . g)
+
 #if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 708
 -- | A modifier for testing functions.
 --
