@@ -3,6 +3,10 @@
 {-# LANGUAGE Rank2Types #-}
 #endif
 -- | Test case generation.
+--
+-- __Note__: the contents of this module (except for the definition of
+-- 'Gen') are re-exported by "Test.QuickCheck". You probably do not
+-- need to import it directly.
 module Test.QuickCheck.Gen where
 
 --------------------------------------------------------------------------
@@ -21,6 +25,9 @@ import Control.Monad
   , filterM
   )
 
+import Control.Monad.Fix
+  ( MonadFix(..) )
+
 import Control.Applicative
   ( Applicative(..) )
 
@@ -62,6 +69,12 @@ instance Monad Gen where
           in m' r2 n
     )
 
+instance MonadFix Gen where
+  mfix f =
+    MkGen $ \r n ->
+      let a = unGen (f a) r n
+      in a
+
 --------------------------------------------------------------------------
 -- ** Primitive generator combinators
 
@@ -83,7 +96,7 @@ variant k (MkGen g) = MkGen (\r n -> g (variantQCGen k r) n)
 sized :: (Int -> Gen a) -> Gen a
 sized f = MkGen (\r n -> let MkGen m = f n in m r n)
 
--- | Generates the size parameter. Used to construct generators that depend on
+-- | Returns the size parameter. Used to construct generators that depend on
 -- the size parameter.
 --
 -- For example, 'listOf', which uses the size parameter as an upper bound on
@@ -156,11 +169,13 @@ gen `suchThatMap` f =
 -- | Tries to generate a value that satisfies a predicate.
 -- If it fails to do so after enough attempts, returns @Nothing@.
 suchThatMaybe :: Gen a -> (a -> Bool) -> Gen (Maybe a)
-gen `suchThatMaybe` p = sized (try 0 . max 1)
+gen `suchThatMaybe` p = sized (\n -> try n (2*n))
  where
-  try _ 0 = return Nothing
-  try k n = do x <- resize (2*k+n) gen
-               if p x then return (Just x) else try (k+1) (n-1)
+  try m n
+    | m > n = return Nothing
+    | otherwise = do
+        x <- resize m gen
+        if p x then return (Just x) else try (m+1) n
 
 -- | Randomly uses one of the given generators. The input list
 -- must be non-empty.
