@@ -17,6 +17,9 @@
 -- They can be pattern-matched on in properties as a stylistic
 -- alternative to using explicit quantification.
 --
+-- __Note__: the contents of this module are re-exported by
+-- "Test.QuickCheck". You do not need to import it directly.
+--
 -- Examples:
 --
 -- @
@@ -49,6 +52,7 @@ module Test.QuickCheck.Modifiers
   , OrderedList(..)
   , NonEmptyList(..)
   , InfiniteList(..)
+  , SortedList(..)
   , Positive(..)
   , NonZero(..)
   , NonNegative(..)
@@ -193,7 +197,7 @@ data InfiniteList a =
 -- the Arbitrary instance generates an infinite list, which is
 -- reduced to a finite prefix by shrinking. We use discard to
 -- check that nothing coming after the finite prefix is used
--- (see makeInfiniteList).
+-- (see infiniteListFromData).
 data InfiniteListInternalData a = Infinite [a] | FinitePrefix [a]
 
 infiniteListFromData :: InfiniteListInternalData a -> InfiniteList a
@@ -222,6 +226,26 @@ instance Arbitrary a => Arbitrary (InfiniteListInternalData a) where
   shrink (FinitePrefix xs) =
     map FinitePrefix (shrink xs)
 
+--------------------------------------------------------------------------
+-- | @Sorted xs@: guarantees that xs is sorted.
+newtype SortedList a = Sorted {getSorted :: [a]}
+ deriving ( Eq, Ord, Show, Read
+#ifndef NO_TYPEABLE
+          , Typeable
+#endif
+          )
+
+instance Functor SortedList where
+  fmap f (Sorted x) = Sorted (map f x)
+
+instance (Arbitrary a, Ord a) => Arbitrary (SortedList a) where
+  arbitrary = fmap (Sorted . sort) arbitrary
+
+  shrink (Sorted xs) =
+    [ Sorted xs'
+    | xs' <- map sort (shrink xs)
+    ]
+
 --------------------------------------------------------------------------
 -- | @Positive x@: guarantees that @x \> 0@.
 newtype Positive a = Positive {getPositive :: a}
