@@ -1,3 +1,4 @@
+{-# OPTIONS_HADDOCK hide #-}
 -- | Combinators for constructing properties.
 {-# LANGUAGE CPP #-}
 #ifndef NO_TYPEABLE
@@ -16,7 +17,7 @@ import Test.QuickCheck.Gen.Unsafe
 import Test.QuickCheck.Arbitrary
 import Test.QuickCheck.Text( isOneLine, putLine )
 import Test.QuickCheck.Exception
-import Test.QuickCheck.State hiding (labels)
+import Test.QuickCheck.State( State(terminal), Confidence(..) )
 
 #ifndef NO_TIMEOUT
 import System.Timeout(timeout)
@@ -34,6 +35,7 @@ import Control.DeepSeq
 #ifndef NO_TYPEABLE
 import Data.Typeable (Typeable)
 #endif
+import Data.Maybe
 
 --------------------------------------------------------------------------
 -- fixities
@@ -94,6 +96,12 @@ class Testable prop where
 
 -- | If a property returns 'Discard', the current test case is discarded,
 -- the same as if a precondition was false.
+--
+-- An example is the definition of '==>':
+--
+-- > (==>) :: Testable prop => Bool -> prop -> Property
+-- > False ==> _ = property Discard
+-- > True  ==> p = property p
 data Discard = Discard
 
 instance Testable Discard where
@@ -131,11 +139,22 @@ morallyDubiousIOProperty = ioProperty
 --
 -- Warning: any random values generated inside of the argument to @ioProperty@
 -- will not currently be shrunk. For best results, generate all random values
--- before calling @ioProperty@.
+-- before calling @ioProperty@, or use 'idempotentIOProperty' if that is safe.
+--
+-- Note: if your property does no quantification, it will only be tested once.
+-- To test it repeatedly, use 'again'.
 ioProperty :: Testable prop => IO prop -> Property
-ioProperty =
+ioProperty prop = idempotentIOProperty (fmap noShrinking prop)
+
+-- | Do I/O inside a property.
+--
+-- Warning: during shrinking, the I/O may not always be re-executed.
+-- Instead, the I/O may be executed once and then its result retained.
+-- If this is not acceptable, use 'ioProperty' instead.
+idempotentIOProperty :: Testable prop => IO prop -> Property
+idempotentIOProperty =
   MkProperty . fmap (MkProp . ioRose . fmap unProp) .
-  promote . fmap (unProperty . noShrinking)
+  promote . fmap (unProperty . property)
 
 instance (Arbitrary a, Show a, Testable prop) => Testable (a -> prop) where
   property f = forAllShrink arbitrary shrink f
@@ -216,16 +235,32 @@ data CallbackKind = Counterexample    -- ^ Affected by the 'verbose' combinator
 -- | The result of a single test.
 data Result
   = MkResult
-  { ok            :: Maybe Bool        -- ^ result of the test case; Nothing = discard
-  , expect        :: Bool              -- ^ indicates what the expected result of the property is
-  , reason        :: String            -- ^ a message indicating what went wrong
-  , theException  :: Maybe AnException -- ^ the exception thrown, if any
-  , abort         :: Bool              -- ^ if True, the test should not be repeated
-  , maybeNumTests :: Maybe Int         -- ^ stop after this many tests
-  , labels        :: Map String Int    -- ^ all labels used by this property
-  , stamp         :: Set String        -- ^ the collected labels for this test case
-  , callbacks     :: [Callback]        -- ^ the callbacks for this test case
-  , testCase      :: [String]          -- ^ the generated test case
+  { ok                 :: Maybe Bool
+    -- ^ result of the test case; Nothing = discard
+  , expect             :: Bool
+    -- ^ indicates what the expected result of the property is
+  , reason             :: String
+    -- ^ a message indicating what went wrong
+  , theException       :: Maybe AnException
+    -- ^ the exception thrown, if any
+  , abort              :: Bool
+    -- ^ if True, the test should not be repeated
+  , maybeNumTests      :: Maybe Int
+    -- ^ stop after this many tests
+  , maybeCheckCoverage :: Maybe Confidence
+    -- ^ required coverage confidence
+  , labels             :: [String]
+    -- ^ test case labels
+  , classes            :: [String]
+    -- ^ test case classes
+  , tables             :: [(String, String)]
+    -- ^ test case tables
+  , requiredCoverage   :: [(Maybe String, String, Double)]
+    -- ^ required coverage
+  , callbacks          :: [Callback]
+    -- ^ the callbacks for this test case
+  , testCase           :: [String]
+    -- ^ the generated test case
   }
 
 exception :: String -> AnException -> Result
@@ -250,16 +285,19 @@ succeeded, failed, rejected :: Result
   where
     result =
       MkResult
-      { ok            = undefined
-      , expect        = True
-      , reason        = ""
-      , theException  = Nothing
-      , abort         = True
-      , maybeNumTests = Nothing
-      , labels        = Map.empty
-      , stamp         = Set.empty
-      , callbacks     = []
-      , testCase      = []
+      { ok                 = undefined
+      , expect             = True
+      , reason             = ""
+      , theException       = Nothing
+      , abort              = True
+      , maybeNumTests      = Nothing
+      , maybeCheckCoverage = Nothing
+      , labels             = []
+      , classes            = []
+      , tables             = []
+      , requiredCoverage   = []
+      , callbacks          = []
+      , testCase           = []
       }
 
 --------------------------------------------------------------------------
@@ -285,9 +323,10 @@ mapProp f = MkProperty . fmap f . unProperty . property
 --------------------------------------------------------------------------
 -- ** Property combinators
 
--- | Changes the maximum test case size for a property.
+-- | Adjust the test case size for a property, by transforming it with the given
+-- function.
 mapSize :: Testable prop => (Int -> Int) -> prop -> Property
-mapSize f p = MkProperty (sized ((`resize` unProperty (property p)) . f))
+mapSize f = property . scale f . unProperty . property
 
 -- | Shrinks the argument to a property if it fails. Shrinking is done
 -- automatically for most types. This function is only needed when you want to
@@ -302,6 +341,7 @@ shrinking shrinker x0 pf = MkProperty (fmap (MkProp . joinRose . fmap unProp) (p
     MkRose (unProperty (property (pf x))) [ props x' | x' <- shrinker x ]
 
 -- | Disables shrinking for a property altogether.
+-- Only quantification /inside/ the call to 'noShrinking' is affected.
 noShrinking :: Testable prop => prop -> Property
 noShrinking = mapRoseResult (onRose (\res _ -> MkRose res []))
 
@@ -353,15 +393,27 @@ whenFail' m =
 -- | Prints out the generated testcase every time the property is tested.
 -- Only variables quantified over /inside/ the 'verbose' are printed.
 verbose :: Testable prop => prop -> Property
-verbose = mapResult (\res -> res { callbacks = newCallbacks (callbacks res) ++ callbacks res })
-  where newCallbacks cbs =
-          PostTest Counterexample (\st res -> putLine (terminal st) (status res ++ ":")):
-          [ PostTest Counterexample f | PostFinalFailure Counterexample f <- cbs ] ++
-          [ PostTest Counterexample (\st res -> putLine (terminal st) "") ]
+verbose = mapResult (\res -> res { callbacks = newCallback (callbacks res):callbacks res })
+  where newCallback cbs =
+          PostTest Counterexample $ \st res -> do
+            putLine (terminal st) (status res ++ ":")
+            sequence_ [ f st res | PostFinalFailure Counterexample f <- cbs ]
+            putLine (terminal st) ""
         status MkResult{ok = Just True} = "Passed"
         status MkResult{ok = Just False} = "Failed"
         status MkResult{ok = Nothing} = "Skipped (precondition false)"
 
+-- | Prints out the generated testcase every time the property fails, including during shrinking.
+-- Only variables quantified over /inside/ the 'verboseShrinking' are printed.
+verboseShrinking :: Testable prop => prop -> Property
+verboseShrinking = mapResult (\res -> res { callbacks = newCallback (callbacks res):callbacks res })
+  where newCallback cbs =
+          PostTest Counterexample $ \st res ->
+            when (ok res == Just False) $ do
+              putLine (terminal st) "Failed:"
+              sequence_ [ f st res | PostFinalFailure Counterexample f <- cbs ]
+              putLine (terminal st) ""
+
 -- | Indicates that a property is supposed to fail.
 -- QuickCheck will report an error if it does not fail.
 expectFailure :: Testable prop => prop -> Property
@@ -387,7 +439,50 @@ again = mapTotalResult (\res -> res{ abort = False })
 withMaxSuccess :: Testable prop => Int -> prop -> Property
 withMaxSuccess n = n `seq` mapTotalResult (\res -> res{ maybeNumTests = Just n })
 
--- | Attaches a label to a property. This is used for reporting
+-- | Check that all coverage requirements defined by 'cover' and 'coverTable'
+-- are met, using a statistically sound test, and fail if they are not met.
+--
+-- Ordinarily, a failed coverage check does not cause the property to fail.
+-- This is because the coverage requirement is not tested in a statistically
+-- sound way. If you use 'cover' to express that a certain value must appear 20%
+-- of the time, QuickCheck will warn you if the value only appears in 19 out of
+-- 100 test cases - but since the coverage varies randomly, you may have just
+-- been unlucky, and there may not be any real problem with your test
+-- generation.
+--
+-- When you use 'checkCoverage', QuickCheck uses a statistical test to account
+-- for the role of luck in coverage failures. It will run as many tests as
+-- needed until it is sure about whether the coverage requirements are met. If a
+-- coverage requirement is not met, the property fails.
+--
+-- Example:
+--
+-- > quickCheck (checkCoverage prop_foo)
+checkCoverage :: Testable prop => prop -> Property
+checkCoverage = checkCoverageWith stdConfidence
+
+-- | Check coverage requirements using a custom confidence level.
+-- See 'stdConfidence'.
+--
+-- An example of making the statistical test less stringent in order to improve
+-- performance:
+--
+-- > quickCheck (checkCoverageWith stdConfidence{certainty = 10^6} prop_foo)
+checkCoverageWith :: Testable prop => Confidence -> prop -> Property
+checkCoverageWith confidence =
+  certainty confidence `seq`
+  tolerance confidence `seq`
+  mapTotalResult (\res -> res{ maybeCheckCoverage = Just confidence })
+
+-- | The standard parameters used by 'checkCoverage': @certainty = 10^9@,
+-- @tolerance = 0.9@. See 'Confidence' for the meaning of the parameters.
+stdConfidence :: Confidence
+stdConfidence =
+  Confidence {
+    certainty = 10^9,
+    tolerance = 0.9 }
+
+-- | Attaches a label to a test case. This is used for reporting
 -- test case distribution.
 --
 -- For example:
@@ -404,10 +499,17 @@ withMaxSuccess n = n `seq` mapTotalResult (\res -> res{ maybeNumTests = Just n }
 -- 5% length of input is 4
 -- 4% length of input is 6
 -- ...
+--
+-- Each use of 'label' in your property results in a separate
+-- table of test case distribution in the output. If this is
+-- not what you want, use 'tabulate'.
 label :: Testable prop => String -> prop -> Property
-label s = classify True s
+label s =
+  s `deepseq`
+  mapTotalResult $
+    \res -> res { labels = s:labels res }
 
--- | Attaches a label to a property. This is used for reporting
+-- | Attaches a label to a test case. This is used for reporting
 -- test case distribution.
 --
 -- > collect x = label (show x)
@@ -426,10 +528,14 @@ label s = classify True s
 -- 5% 4
 -- 4% 6
 -- ...
+--
+-- Each use of 'collect' in your property results in a separate
+-- table of test case distribution in the output. If this is
+-- not what you want, use 'tabulate'.
 collect :: (Show a, Testable prop) => a -> prop -> Property
 collect x = label (show x)
 
--- | Records how many test cases satisfy a given condition.
+-- | Reports how many test cases satisfy a given condition.
 --
 -- For example:
 --
@@ -445,39 +551,170 @@ classify :: Testable prop =>
             Bool    -- ^ @True@ if the test case should be labelled.
          -> String  -- ^ Label.
          -> prop -> Property
-classify b s = cover b 0 s
+classify False _ = property
+classify True s =
+  s `deepseq`
+  mapTotalResult $
+    \res -> res { classes = s:classes res }
 
 -- | Checks that at least the given proportion of /successful/ test
 -- cases belong to the given class. Discarded tests (i.e. ones
 -- with a false precondition) do not affect coverage.
 --
+-- __Note:__ If the coverage check fails, QuickCheck prints out a warning, but
+-- the property does /not/ fail. To make the property fail, use 'checkCoverage'.
+--
 -- For example:
 --
 -- > prop_sorted_sort :: [Int] -> Property
 -- > prop_sorted_sort xs =
 -- >   sorted xs ==>
--- >   cover (length xs > 1) 50 "non-trivial" $
+-- >   cover 50 (length xs > 1) "non-trivial" $
 -- >   sort xs === xs
 --
 -- >>> quickCheck prop_sorted_sort
--- *** Insufficient coverage after 100 tests (only 24% non-trivial, not 50%).
+-- +++ OK, passed 100 tests; 135 discarded (26% non-trivial).
+-- <BLANKLINE>
+-- Only 26% non-trivial, but expected 50%
 cover :: Testable prop =>
-         Bool   -- ^ @True@ if the test case belongs to the class.
-      -> Int    -- ^ The required percentage (0-100) of test cases.
+         Double -- ^ The required percentage (0-100) of test cases.
+      -> Bool   -- ^ @True@ if the test case belongs to the class.
       -> String -- ^ Label for the test case class.
       -> prop -> Property
-cover x n s =
-  x `seq` n `seq` s `listSeq`
+cover p x s = mapTotalResult f . classify x s
+  where
+    f res = res { requiredCoverage = (Nothing, s, p/100):requiredCoverage res }
+
+-- | Collects information about test case distribution into a table.
+-- The arguments to 'tabulate' are the table's name and a list of values
+-- associated with the current test case. After testing, QuickCheck prints the
+-- frequency of all collected values. The frequencies are expressed as a
+-- percentage of the total number of values collected.
+--
+-- You should prefer 'tabulate' to 'label' when each test case is associated
+-- with a varying number of values. Here is a (not terribly useful) example,
+-- where the test data is a list of integers and we record all values that
+-- occur in the list:
+--
+-- > prop_sorted_sort :: [Int] -> Property
+-- > prop_sorted_sort xs =
+-- >   sorted xs ==>
+-- >   tabulate "List elements" (map show xs) $
+-- >   sort xs === xs
+--
+-- >>> quickCheck prop_sorted_sort
+-- +++ OK, passed 100 tests; 1684 discarded.
+-- <BLANKLINE>
+-- List elements (109 in total):
+--  3.7% 0
+--  3.7% 17
+--  3.7% 2
+--  3.7% 6
+--  2.8% -6
+--  2.8% -7
+--
+-- Here is a more useful example. We are testing a chatroom, where the user can
+-- log in, log out, or send a message:
+--
+-- > data Command = LogIn | LogOut | SendMessage String deriving (Data, Show)
+-- > instance Arbitrary Command where ...
+--
+-- There are some restrictions on command sequences; for example, the user must
+-- log in before doing anything else. The function @valid :: [Command] -> Bool@
+-- checks that a command sequence is allowed. Our property then has the form:
+--
+-- > prop_chatroom :: [Command] -> Property
+-- > prop_chatroom cmds =
+-- >   valid cmds ==>
+-- >     ...
+--
+-- The use of '==>' may skew test case distribution. We use 'collect' to see the
+-- length of the command sequences, and 'tabulate' to get the frequencies of the
+-- individual commands:
+--
+-- > prop_chatroom :: [Command] -> Property
+-- > prop_chatroom cmds =
+-- >   wellFormed cmds LoggedOut ==>
+-- >   'collect' (length cmds) $
+-- >   'tabulate' "Commands" (map (show . 'Data.Data.toConstr') cmds) $
+-- >     ...
+--
+-- >>> quickCheckWith stdArgs{maxDiscardRatio = 1000} prop_chatroom
+-- +++ OK, passed 100 tests; 2775 discarded:
+-- 60% 0
+-- 20% 1
+-- 15% 2
+--  3% 3
+--  1% 4
+--  1% 5
+-- <BLANKLINE>
+-- Commands (68 in total):
+-- 62% LogIn
+-- 22% SendMessage
+-- 16% LogOut
+tabulate :: Testable prop => String -> [String] -> prop -> Property
+tabulate key values =
+  key `deepseq` values `deepseq`
   mapTotalResult $
-    \res -> res {
-      labels = Map.insertWith max s n (labels res),
-      stamp = if x then Set.insert s (stamp res) else stamp res }
-  where [] `listSeq` z = z
-        (x:xs) `listSeq` z = x `seq` xs `listSeq` z
+    \res -> res { tables = [(key, value) | value <- values] ++ tables res }
+
+-- | Checks that the values in a given 'table' appear a certain proportion of
+-- the time. A call to 'coverTable' @table@ @[(x1, p1), ..., (xn, pn)]@ asserts
+-- that of the values in @table@, @x1@ should appear at least @p1@ percent of
+-- the time, @x2@ at least @p2@ percent of the time, and so on.
+--
+-- __Note:__ If the coverage check fails, QuickCheck prints out a warning, but
+-- the property does /not/ fail. To make the property fail, use 'checkCoverage'.
+--
+-- Continuing the example from the 'tabular' combinator...
+--
+-- > data Command = LogIn | LogOut | SendMessage String deriving (Data, Show)
+-- > prop_chatroom :: [Command] -> Property
+-- > prop_chatroom cmds =
+-- >   wellFormed cmds LoggedOut ==>
+-- >   'tabulate' "Commands" (map (show . 'Data.Data.toConstr') cmds) $
+-- >     ...
+--
+-- ...we can add a coverage requirement as follows, which checks that @LogIn@,
+-- @LogOut@ and @SendMessage@ each occur at least 25% of the time:
+--
+-- > prop_chatroom :: [Command] -> Property
+-- > prop_chatroom cmds =
+-- >   wellFormed cmds LoggedOut ==>
+-- >   coverTable "Commands" [("LogIn", 25), ("LogOut", 25), ("SendMessage", 25)] $
+-- >   'tabulate' "Commands" (map (show . 'Data.Data.toConstr') cmds) $
+-- >     ... property goes here ...
+--
+-- >>> quickCheck prop_chatroom
+-- +++ OK, passed 100 tests; 2909 discarded:
+-- 56% 0
+-- 17% 1
+-- 10% 2
+--  6% 3
+--  5% 4
+--  3% 5
+--  3% 7
+-- <BLANKLINE>
+-- Commands (111 in total):
+-- 51.4% LogIn
+-- 30.6% SendMessage
+-- 18.0% LogOut
+-- <BLANKLINE>
+-- Table 'Commands' had only 18.0% LogOut, but expected 25.0%
+coverTable :: Testable prop =>
+  String -> [(String, Double)] -> prop -> Property
+coverTable table xs =
+  tables `deepseq` xs `deepseq`
+  mapTotalResult $
+    \res -> res { requiredCoverage = ys ++ requiredCoverage res }
+  where
+    ys = [(Just table, x, p/100) | (x, p) <- xs]
 
 -- | Implication for properties: The resulting property holds if
 -- the first argument is 'False' (in which case the test case is discarded),
--- or if the given property holds.
+-- or if the given property holds. Note that using implication carelessly can
+-- severely skew test case distribution: consider using 'cover' to make sure
+-- that your test data is still good quality.
 (==>) :: Testable prop => Bool -> prop -> Property
 False ==> _ = property Discard
 True  ==> p = property p
@@ -506,16 +743,38 @@ forAll :: (Show a, Testable prop)
        => Gen a -> (a -> prop) -> Property
 forAll gen pf = forAllShrink gen (\_ -> []) pf
 
+-- | Like 'forAll', but with an explicitly given show function.
+forAllShow :: Testable prop
+           => Gen a -> (a -> String) -> (a -> prop) -> Property
+forAllShow gen shower pf = forAllShrinkShow gen (\_ -> []) shower pf
+
+-- | Like 'forAll', but without printing the generated value.
+forAllBlind :: Testable prop
+           => Gen a -> (a -> prop) -> Property
+forAllBlind gen pf = forAllShrinkBlind gen (\_ -> []) pf
+
 -- | Like 'forAll', but tries to shrink the argument for failing test cases.
 forAllShrink :: (Show a, Testable prop)
              => Gen a -> (a -> [a]) -> (a -> prop) -> Property
-forAllShrink gen shrinker pf =
+forAllShrink gen shrinker = forAllShrinkShow gen shrinker show
+
+-- | Like 'forAllShrink', but with an explicitly given show function.
+forAllShrinkShow
+  :: Testable prop
+  => Gen a -> (a -> [a]) -> (a -> String) -> (a -> prop) -> Property
+forAllShrinkShow gen shrinker shower pf =
+  forAllShrinkBlind gen shrinker (\x -> counterexample (shower x) (pf x))
+
+-- | Like 'forAllShrink', but without printing the generated value.
+forAllShrinkBlind
+  :: Testable prop
+  => Gen a -> (a -> [a]) -> (a -> prop) -> Property
+forAllShrinkBlind gen shrinker pf =
   again $
   MkProperty $
   gen >>= \x ->
     unProperty $
-    shrinking shrinker x $ \x' ->
-      counterexample (show x') (pf x')
+    shrinking shrinker x pf
 
 -- | Nondeterministic choice: 'p1' '.&.' 'p2' picks randomly one of
 -- 'p1' and 'p2' to test. If you test the property 100 times it
@@ -549,10 +808,10 @@ conjoin ps =
     case ok result of
       _ | not (expect result) ->
         return (return failed { reason = "expectFailure may not occur inside a conjunction" })
-      Just True -> return (conj (addLabels result . addCallbacks result . k) ps)
+      Just True -> return (conj (addLabels result . addCallbacksAndCoverage result . k) ps)
       Just False -> return rose
       Nothing -> do
-        rose2@(MkRose result2 _) <- reduceRose (conj (addCallbacks result . k) ps)
+        rose2@(MkRose result2 _) <- reduceRose (conj (addCallbacksAndCoverage result . k) ps)
         return $
           -- Nasty work to make sure we use the right callbacks
           case ok result2 of
@@ -560,11 +819,13 @@ conjoin ps =
             Just False -> rose2
             Nothing -> rose2
 
-  addCallbacks result r =
-    r { callbacks = callbacks result ++ callbacks r }
+  addCallbacksAndCoverage result r =
+    r { callbacks = callbacks result ++ callbacks r,
+        requiredCoverage = requiredCoverage result ++ requiredCoverage r }
   addLabels result r =
-    r { labels = Map.unionWith max (labels result) (labels r),
-        stamp = Set.union (stamp result) (stamp r) }
+    r { labels = labels result ++ labels r,
+        classes = classes result ++ classes r,
+        tables = tables result ++ tables r }
 
 -- | Disjunction: 'p1' '.||.' 'p2' passes unless 'p1' and 'p2' simultaneously fail.
 (.||.) :: (Testable prop1, Testable prop2) => prop1 -> prop2 -> Property
@@ -583,25 +844,27 @@ disjoin ps =
     do result1 <- p
        case ok result1 of
          _ | not (expect result1) -> return expectFailureError
-         Just True -> return result1
          Just False -> do
            result2 <- q
            return $
              case ok result2 of
                _ | not (expect result2) -> expectFailureError
-               Just True -> result2
+               Just True -> addCoverage result1 result2
                Just False ->
                  MkResult {
                    ok = Just False,
                    expect = True,
                    reason = sep (reason result1) (reason result2),
                    theException = theException result1 `mplus` theException result2,
-                   -- The following three fields are not important because the
+                   -- The following few fields are not important because the
                    -- test case has failed anyway
                    abort = False,
                    maybeNumTests = Nothing,
-                   labels = Map.empty,
-                   stamp = Set.empty,
+                   maybeCheckCoverage = Nothing,
+                   labels = [],
+                   classes = [],
+                   tables = [],
+                   requiredCoverage = [],
                    callbacks =
                      callbacks result1 ++
                      [PostFinalFailure Counterexample $ \st _res -> putLine (terminal st) ""] ++
@@ -610,18 +873,28 @@ disjoin ps =
                      testCase result1 ++
                      testCase result2 }
                Nothing -> result2
-         Nothing -> do
-           result2 <- q
-           return (case ok result2 of
-                     _ | not (expect result2) -> expectFailureError
-                     Just True -> result2
-                     _ -> result1)
+         -- The "obvious" semantics of .||. has:
+         --   discard .||. true = true
+         --   discard .||. discard = discard
+         -- but this implementation gives discard .||. true = discard.
+         -- This is reasonable because evaluating result2 in the case
+         -- that result1 discards is just busy-work - it won't ever
+         -- cause the property to fail. On the other hand, discarding
+         -- instead of returning true causes us to execute one more
+         -- test case - but assuming that preconditions are cheap to
+         -- evaluate, this is no more work than evaluating result2
+         -- would be, while (unlike evaluating result2) it might catch
+         -- a bug.
+         _ -> return result1
 
   expectFailureError = failed { reason = "expectFailure may not occur inside a disjunction" }
   sep [] s = s
   sep s [] = s
   sep s s' = s ++ ", " ++ s'
 
+  addCoverage result r =
+    r { requiredCoverage = requiredCoverage result ++ requiredCoverage r }
+
 -- | Like '==', but prints a counterexample when it fails.
 infix 4 ===
 (===) :: (Eq a, Show a) => a -> a -> Property
@@ -632,6 +905,16 @@ x === y =
     interpret True  = " == "
     interpret False = " /= "
 
+-- | Like '/=', but prints a counterexample when it fails.
+infix 4 =/=
+(=/=) :: (Eq a, Show a) => a -> a -> Property
+x =/= y =
+  counterexample (show x ++ interpret res ++ show y) res
+  where
+    res = x /= y
+    interpret True  = " /= "
+    interpret False = " == "
+
 #ifndef NO_DEEPSEQ
 -- | Checks that a value is total, i.e., doesn't crash when evaluated.
 total :: NFData a => a -> Property
