@@ -1,7 +1,8 @@
+{-# OPTIONS_HADDOCK hide #-}
 -- | The main test loop.
 {-# LANGUAGE CPP #-}
 #ifndef NO_SAFE_HASKELL
-{-# LANGUAGE Safe #-}
+{-# LANGUAGE Trustworthy #-}
 #endif
 module Test.QuickCheck.Test where
 
@@ -9,13 +10,14 @@ module Test.QuickCheck.Test where
 -- imports
 
 import Test.QuickCheck.Gen
-import Test.QuickCheck.Property hiding ( Result( reason, theException, labels ) )
+import Test.QuickCheck.Property hiding ( Result( reason, theException, labels, classes, tables ), (.&.) )
 import qualified Test.QuickCheck.Property as P
 import Test.QuickCheck.Text
-import Test.QuickCheck.State hiding (labels)
+import Test.QuickCheck.State hiding (labels, classes, tables, requiredCoverage)
 import qualified Test.QuickCheck.State as S
 import Test.QuickCheck.Exception
 import Test.QuickCheck.Random
+import Data.Number.Erf(invnormcdf)
 import System.Random(split)
 #if defined(MIN_VERSION_containers)
 #if MIN_VERSION_containers(0,5,0)
@@ -27,6 +29,8 @@ import qualified Data.Map as Map
 import qualified Data.Map as Map
 #endif
 import qualified Data.Set as Set
+import Data.Set(Set)
+import Data.Map(Map)
 
 import Data.Char
   ( isSpace
@@ -37,11 +41,15 @@ import Data.List
   , sortBy
   , group
   , intersperse
+  , intercalate
   )
 
-import Data.Maybe(fromMaybe)
+import Data.Maybe(fromMaybe, isNothing, catMaybes)
 import Data.Ord(comparing)
 import Text.Printf(printf)
+import Data.Either(lefts, rights)
+import Control.Monad
+import Data.Bits
 
 --------------------------------------------------------------------------
 -- quickCheck
@@ -77,41 +85,65 @@ data Args
 data Result
   -- | A successful test run
   = Success
-    { numTests       :: Int               -- ^ Number of tests performed
-    , labels         :: [(String,Double)] -- ^ Labels and frequencies found during all successful tests
-    , output         :: String            -- ^ Printed output
+    { numTests     :: Int
+      -- ^ Number of tests performed
+    , numDiscarded :: Int
+      -- ^ Number of tests skipped
+    , labels       :: !(Map [String] Int)
+      -- ^ The number of test cases having each combination of labels (see 'label')
+    , classes      :: !(Map String Int)
+      -- ^ The number of test cases having each class (see 'classify')
+    , tables       :: !(Map String (Map String Int))
+      -- ^ Data collected by 'tabulate'
+    , output       :: String
+      -- ^ Printed output
     }
   -- | Given up
   | GaveUp
-    { numTests       :: Int               --   Number of tests performed
-    , labels         :: [(String,Double)] --   Labels and frequencies found during all successful tests
-    , output         :: String            --   Printed output
+    { numTests     :: Int
+    , numDiscarded :: Int
+      -- ^ Number of tests skipped
+    , labels       :: !(Map [String] Int)
+    , classes      :: !(Map String Int)
+    , tables       :: !(Map String (Map String Int))
+    , output       :: String
     }
   -- | A failed test run
   | Failure
-    { numTests        :: Int               --   Number of tests performed
-    , numShrinks      :: Int               -- ^ Number of successful shrinking steps performed
-    , numShrinkTries  :: Int               -- ^ Number of unsuccessful shrinking steps performed
-    , numShrinkFinal  :: Int               -- ^ Number of unsuccessful shrinking steps performed since last successful shrink
-    , usedSeed        :: QCGen             -- ^ What seed was used
-    , usedSize        :: Int               -- ^ What was the test size
-    , reason          :: String            -- ^ Why did the property fail
-    , theException    :: Maybe AnException -- ^ The exception the property threw, if any
-    , labels          :: [(String,Double)] --   Labels and frequencies found during all successful tests
-    , output          :: String            --   Printed output
-    , failingTestCase :: [String]          -- ^ The test case which provoked the failure
+    { numTests        :: Int
+    , numDiscarded    :: Int
+      -- ^ Number of tests skipped
+    , numShrinks      :: Int
+      -- ^ Number of successful shrinking steps performed
+    , numShrinkTries  :: Int
+      -- ^ Number of unsuccessful shrinking steps performed
+    , numShrinkFinal  :: Int
+      -- ^ Number of unsuccessful shrinking steps performed since last successful shrink
+    , usedSeed        :: QCGen
+      -- ^ What seed was used
+    , usedSize        :: Int
+      -- ^ What was the test size
+    , reason          :: String
+      -- ^ Why did the property fail
+    , theException    :: Maybe AnException
+      -- ^ The exception the property threw, if any
+    , output          :: String
+    , failingTestCase :: [String]
+      -- ^ The test case which provoked the failure
+    , failingLabels   :: [String]
+      -- ^ The test case's labels (see 'label')
+    , failingClasses  :: Set String
+      -- ^ The test case's classes (see 'classify')
     }
   -- | A property that should have failed did not
   | NoExpectedFailure
-    { numTests       :: Int               --   Number of tests performed
-    , labels         :: [(String,Double)] --   Labels and frequencies found during all successful tests
-    , output         :: String            --   Printed output
-    }
- -- | The tests passed but a use of 'cover' had insufficient coverage
- | InsufficientCoverage
-    { numTests       :: Int               --   Number of tests performed
-    , labels         :: [(String,Double)] --   Labels and frequencies found during all successful tests
-    , output         :: String            --   Printed output
+    { numTests     :: Int
+    , numDiscarded :: Int
+      -- ^ Number of tests skipped
+    , labels       :: !(Map [String] Int)
+    , classes      :: !(Map String Int)
+    , tables       :: !(Map String (Map String Int))
+    , output       :: String
     }
  deriving ( Show )
 
@@ -148,12 +180,17 @@ quickCheckResult p = quickCheckWithResult stdArgs p
 
 -- | Tests a property, using test arguments, produces a test result, and prints the results to 'stdout'.
 quickCheckWithResult :: Testable prop => Args -> prop -> IO Result
-quickCheckWithResult a p = (if chatty a then withStdioTerminal else withNullTerminal) $ \tm -> do
+quickCheckWithResult a p =
+  withState a (\s -> test s (property p))
+
+withState :: Args -> (State -> IO a) -> IO a
+withState a test = (if chatty a then withStdioTerminal else withNullTerminal) $ \tm -> do
      rnd <- case replay a of
               Nothing      -> newQCGen
               Just (rnd,_) -> return rnd
      test MkState{ terminal                  = tm
                  , maxSuccessTests           = maxSuccess a
+                 , coverageConfidence        = Nothing
                  , maxDiscardedRatio         = maxDiscardRatio a
                  , computeSize               = case replay a of
                                                  Nothing    -> computeSize'
@@ -163,13 +200,15 @@ quickCheckWithResult a p = (if chatty a then withStdioTerminal else withNullTerm
                  , numDiscardedTests         = 0
                  , numRecentlyDiscardedTests = 0
                  , S.labels                  = Map.empty
-                 , collected                 = []
-                 , expectedFailure           = False
+                 , S.classes                 = Map.empty
+                 , S.tables                  = Map.empty
+                 , S.requiredCoverage        = Map.empty
+                 , expected                  = True
                  , randomSeed                = rnd
                  , numSuccessShrinks         = 0
                  , numTryShrinks             = 0
                  , numTotTryShrinks          = 0
-                 } (unGen (unProperty (property p)))
+                 }
   where computeSize' n d
           -- e.g. with maxSuccess = 250, maxSize = 100, goes like this:
           -- 0, 1, 2, ..., 99, 0, 1, 2, ..., 99, 0, 2, 4, ..., 98.
@@ -205,124 +244,137 @@ verboseCheckWithResult a p = quickCheckWithResult a (verbose p)
 --------------------------------------------------------------------------
 -- main test loop
 
-test :: State -> (QCGen -> Int -> Prop) -> IO Result
+test :: State -> Property -> IO Result
 test st f
-  | numSuccessTests st   >= maxSuccessTests st =
+  | numSuccessTests st   >= maxSuccessTests st && isNothing (coverageConfidence st) =
     doneTesting st f
-  | numDiscardedTests st >= maxDiscardedRatio st * maxSuccessTests st =
+  | numDiscardedTests st >= maxDiscardedRatio st * max (numSuccessTests st) (maxSuccessTests st) =
     giveUp st f
   | otherwise =
     runATest st f
 
-doneTesting :: State -> (QCGen -> Int -> Prop) -> IO Result
+doneTesting :: State -> Property -> IO Result
 doneTesting st _f
-  | not (expectedFailure st) = do
+  | expected st == False = do
       putPart (terminal st)
         ( bold ("*** Failed!")
        ++ " Passed "
-       ++ show (numSuccessTests st)
-       ++ " tests (expected failure)"
+       ++ showTestCount st
+       ++ " (expected failure)"
         )
       finished NoExpectedFailure
-  | not (null (insufficientlyCovered st)) = do
-      putPart (terminal st)
-        ( bold ("*** Insufficient coverage after ")
-       ++ show (numSuccessTests st)
-       ++ " tests"
-        )
-      finished InsufficientCoverage
   | otherwise = do
       putPart (terminal st)
         ( "+++ OK, passed "
-       ++ show (numSuccessTests st)
-       ++ " tests"
+       ++ showTestCount st
         )
       finished Success
   where
     finished k = do
       success st
       theOutput <- terminalOutput (terminal st)
-      return (k (numSuccessTests st) (summary st) theOutput)
+      return (k (numSuccessTests st) (numDiscardedTests st) (S.labels st) (S.classes st) (S.tables st) theOutput)
 
-giveUp :: State -> (QCGen -> Int -> Prop) -> IO Result
+giveUp :: State -> Property -> IO Result
 giveUp st _f =
   do -- CALLBACK gave_up?
      putPart (terminal st)
        ( bold ("*** Gave up!")
       ++ " Passed only "
-      ++ show (numSuccessTests st)
+      ++ showTestCount st
       ++ " tests"
        )
      success st
      theOutput <- terminalOutput (terminal st)
-     return GaveUp{ numTests = numSuccessTests st
-                  , labels   = summary st
-                  , output   = theOutput
+     return GaveUp{ numTests     = numSuccessTests st
+                  , numDiscarded = numDiscardedTests st
+                  , labels       = S.labels st
+                  , classes      = S.classes st
+                  , tables       = S.tables st
+                  , output       = theOutput
                   }
 
-runATest :: State -> (QCGen -> Int -> Prop) -> IO Result
+showTestCount :: State -> String
+showTestCount st =
+     number (numSuccessTests st) "test"
+  ++ concat [ "; " ++ show (numDiscardedTests st) ++ " discarded"
+            | numDiscardedTests st > 0
+            ]
+
+runATest :: State -> Property -> IO Result
 runATest st f =
   do -- CALLBACK before_test
      putTemp (terminal st)
         ( "("
-       ++ number (numSuccessTests st) "test"
-       ++ concat [ "; " ++ show (numDiscardedTests st) ++ " discarded"
-                 | numDiscardedTests st > 0
-                 ]
+       ++ showTestCount st
        ++ ")"
         )
+     let powerOfTwo n = n .&. (n - 1) == 0
+     let f_or_cov =
+           case coverageConfidence st of
+             Just confidence | (1 + numSuccessTests st) `mod` 100 == 0 && powerOfTwo ((1 + numSuccessTests st) `div` 100) ->
+               addCoverageCheck confidence st f
+             _ -> f
      let size = computeSize st (numSuccessTests st) (numRecentlyDiscardedTests st)
-     MkRose res ts <- protectRose (reduceRose (unProp (f rnd1 size)))
+     MkRose res ts <- protectRose (reduceRose (unProp (unGen (unProperty f_or_cov) rnd1 size)))
      res <- callbackPostTest st res
 
      let continue break st' | abort res = break st'
                             | otherwise = test st'
-         cons x xs
-           | Set.null x = xs
-           | otherwise = x:xs
+
+     let st' = st{ coverageConfidence = maybeCheckCoverage res `mplus` coverageConfidence st
+                 , maxSuccessTests = fromMaybe (maxSuccessTests st) (maybeNumTests res)
+                 , S.labels = Map.insertWith (+) (P.labels res) 1 (S.labels st)
+                 , S.classes = Map.unionWith (+) (S.classes st) (Map.fromList (zip (P.classes res) (repeat 1)))
+                 , S.tables =
+                   foldr (\(tab, x) -> Map.insertWith (Map.unionWith (+)) tab (Map.singleton x 1))
+                     (S.tables st) (P.tables res)
+                 , S.requiredCoverage =
+                   foldr (\(key, value, p) -> Map.insertWith max (key, value) p)
+                     (S.requiredCoverage st) (P.requiredCoverage res)
+                 , expected = expect res }
 
      case res of
-       MkResult{ok = Just True, stamp = stamp, expect = expect, maybeNumTests = mnt} -> -- successful test
+       MkResult{ok = Just True} -> -- successful test
          do continue doneTesting
-              st{ numSuccessTests           = numSuccessTests st + 1
-                , numRecentlyDiscardedTests = 0
-                , maxSuccessTests           = fromMaybe (maxSuccessTests st) mnt
-                , randomSeed                = rnd2
-                , S.labels                  = Map.unionWith max (S.labels st) (P.labels res)
-                , collected                 = stamp `cons` collected st
-                , expectedFailure           = expect
-                } f
+              st'{ numSuccessTests           = numSuccessTests st' + 1
+                 , numRecentlyDiscardedTests = 0
+                 , randomSeed = rnd2
+                 } f
 
-       MkResult{ok = Nothing, expect = expect, maybeNumTests = mnt} -> -- discarded test
+       MkResult{ok = Nothing, expect = expect, maybeNumTests = mnt, maybeCheckCoverage = mcc} -> -- discarded test
          do continue giveUp
-              st{ numDiscardedTests         = numDiscardedTests st + 1
-                , numRecentlyDiscardedTests = numRecentlyDiscardedTests st + 1
-                , maxSuccessTests           = fromMaybe (maxSuccessTests st) mnt
-                , randomSeed                = rnd2
-                , S.labels                  = Map.unionWith max (S.labels st) (P.labels res)
-                , expectedFailure           = expect
+              -- Don't add coverage info from this test
+              st{ numDiscardedTests         = numDiscardedTests st' + 1
+                , numRecentlyDiscardedTests = numRecentlyDiscardedTests st' + 1
+                , randomSeed = rnd2
                 } f
 
        MkResult{ok = Just False} -> -- failed test
-         do (numShrinks, totFailed, lastFailed, res) <- foundFailure st res ts
-            theOutput <- terminalOutput (terminal st)
+         do (numShrinks, totFailed, lastFailed, res) <- foundFailure st' res ts
+            theOutput <- terminalOutput (terminal st')
             if not (expect res) then
-              return Success{ labels = summary st,
-                              numTests = numSuccessTests st+1,
+              return Success{ labels = S.labels st',
+                              classes = S.classes st',
+                              tables = S.tables st',
+                              numTests = numSuccessTests st'+1,
+                              numDiscarded = numDiscardedTests st',
                               output = theOutput }
              else do
               testCase <- mapM showCounterexample (P.testCase res)
-              return Failure{ usedSeed        = randomSeed st -- correct! (this will be split first)
+              return Failure{ usedSeed        = randomSeed st' -- correct! (this will be split first)
                             , usedSize        = size
-                            , numTests        = numSuccessTests st+1
+                            , numTests        = numSuccessTests st'+1
+                            , numDiscarded    = numDiscardedTests st'
                             , numShrinks      = numShrinks
                             , numShrinkTries  = totFailed
                             , numShrinkFinal  = lastFailed
                             , output          = theOutput
                             , reason          = P.reason res
                             , theException    = P.theException res
-                            , labels          = summary st
                             , failingTestCase = testCase
+                            , failingLabels   = P.labels res
+                            , failingClasses  = Set.fromList (P.classes res)
                             }
  where
   (rnd1,rnd2) = split (randomSeed st)
@@ -367,83 +419,73 @@ failureSummaryAndReason st res = (summary, full)
       where
         showNumTryShrinks = full && numTryShrinks st > 0
 
-summary :: State -> [(String, Double)]
-summary st = reverse
-           . sortBy (comparing snd)
-           . map (\ss -> (head ss, fromIntegral (length ss) * 100 / fromIntegral (numSuccessTests st)))
-           . group
-           . sort
-           $ [ concat (intersperse ", " s')
-             | s <- collected st
-               -- HACK: don't print out labels that were created by 'cover'.
-             , let s' = [ t | t <- Set.toList s, Map.lookup t (S.labels st) == Just 0 ]
-             , not (null s')
-             ]
-
 success :: State -> IO ()
-success st =
-  case allLabels ++ covers of
-    []    -> do putLine (terminal st) "."
-    [pt]  -> do putLine (terminal st)
-                  ( " ("
-                 ++ dropWhile isSpace pt
-                 ++ ")."
-                  )
-    cases -> do putLine (terminal st) ":"
-                mapM_ (putLine $ terminal st) cases
- where
-  allLabels :: [String]
-  allLabels = map (formatLabel (numSuccessTests st) True) (summary st)
+success st = do
+  mapM_ (putLine $ terminal st) (paragraphs [short, long])
+  where
+    (short, long) =
+      case labelsAndTables st of
+        ([msg], long) ->
+          ([" (" ++ dropWhile isSpace msg ++ ")."], long)
+        ([], long) ->
+          (["."], long)
+        (short, long) ->
+          (":":short, long)
 
-  covers :: [String]
-  covers = [ ("only " ++ formatLabel (numSuccessTests st) False (l, p) ++ ", not " ++ show reqP ++ "%")
-           | (l, reqP, p) <- insufficientlyCovered st ]
+labelsAndTables :: State -> ([String], [String])
+labelsAndTables st = (labels, tables)
+  where
+    labels :: [String]
+    labels =
+      paragraphs $
+        [ showTable (numSuccessTests st) Nothing m
+        | m <- S.classes st:Map.elems numberedLabels ]
 
-formatLabel :: Int -> Bool -> (String, Double) -> String
-formatLabel n pad (x, p) = showP pad p ++ " " ++ x
- where
-  showP :: Bool -> Double -> String
-  showP pad p =
-    (if pad && p < 10 then " " else "") ++
-    printf "%.*f" places p ++ "%"
+    numberedLabels :: Map Int (Map String Int)
+    numberedLabels =
+      Map.fromListWith (Map.unionWith (+)) $
+        [ (i, Map.singleton l n)
+        | (labels, n) <- Map.toList (S.labels st),
+          (i, l) <- zip [0..] labels ]
 
-  -- Show no decimal places if <= 100 successful tests,
-  -- one decimal place if <= 1000 successful tests,
-  -- two decimal places if <= 10000 successful tests, and so on.
-  places :: Integer
-  places =
-    ceiling (logBase 10 (fromIntegral n) - 2 :: Double) `max` 0
+    tables :: [String]
+    tables =
+      paragraphs $
+        [ showTable (sum (Map.elems m)) (Just table) m
+        | (table, m) <- Map.toList (S.tables st) ] ++
+        [[ (case mtable of Nothing -> "Only "; Just table -> "Table '" ++ table ++ "' had only ")
+         ++ lpercent n tot ++ " " ++ label ++ ", but expected " ++ lpercentage p tot
+         | (mtable, label, tot, n, p) <- allCoverage st,
+           insufficientlyCovered (fmap certainty (coverageConfidence st)) tot n p ]]
 
-labelCount :: String -> State -> Int
-labelCount l st =
-  -- XXX in case of a disjunction, a label can occur several times,
-  -- need to think what to do there
-  length [ l' | l' <- concat (map Set.toList (collected st)), l == l' ]
+showTable :: Int -> Maybe String -> Map String Int -> [String]
+showTable k mtable m =
+  [table ++ " " ++ total ++ ":" | Just table <- [mtable]] ++
+  (map format .
+   -- Descending order of occurrences
+   reverse . sortBy (comparing snd) .
+   -- If #occurences the same, sort in increasing order of key
+   -- (note: works because sortBy is stable)
+   reverse . sortBy (comparing fst) $ Map.toList m)
+  where
+    format (key, v) =
+      rpercent v k ++ " " ++ key
 
-percentage :: Integral a => State -> a -> Double
-percentage st n =
-  fromIntegral n * 100 / fromIntegral (numSuccessTests st)
-
-insufficientlyCovered :: State -> [(String, Int, Double)]
-insufficientlyCovered st =
-  [ (l, reqP, p)
-  | (l, reqP) <- Map.toList (S.labels st),
-    let p = percentage st (labelCount l st),
-    p < fromIntegral reqP ]
+    total = printf "(%d in total)" k
 
 --------------------------------------------------------------------------
 -- main shrinking loop
 
 foundFailure :: State -> P.Result -> [Rose P.Result] -> IO (Int, Int, Int, P.Result)
 foundFailure st res ts =
-  do localMin st{ numTryShrinks = 0 } res res ts
+  do localMin st{ numTryShrinks = 0 } res ts
 
-localMin :: State -> P.Result -> P.Result -> [Rose P.Result] -> IO (Int, Int, Int, P.Result)
+localMin :: State -> P.Result -> [Rose P.Result] -> IO (Int, Int, Int, P.Result)
 -- Don't try to shrink for too long
-localMin st res _ ts
+localMin st res ts
   | numSuccessShrinks st + numTotTryShrinks st >= numTotMaxShrinks st =
     localMinFound st res
-localMin st res _ ts = do
+localMin st res ts = do
   r <- tryEvaluateIO $
     putTemp (terminal st) (failureSummary st res)
   case r of
@@ -465,9 +507,9 @@ localMin' st res (t:ts) =
     res' <- callbackPostTest st res'
     if ok res' == Just False
       then localMin st{ numSuccessShrinks = numSuccessShrinks st + 1,
-                        numTryShrinks     = 0 } res' res ts'
+                        numTryShrinks     = 0 } res' ts'
       else localMin st{ numTryShrinks    = numTryShrinks st + 1,
-                        numTotTryShrinks = numTotTryShrinks st + 1 } res res ts
+                        numTotTryShrinks = numTotTryShrinks st + 1 } res ts
 
 localMinFound :: State -> P.Result -> IO (Int, Int, Int, P.Result)
 localMinFound st res =
@@ -495,5 +537,72 @@ callbackPostFinalFailure st res = do
       return ()
     Right () -> return ()
 
+----------------------------------------------------------------------
+-- computing coverage
+
+sufficientlyCovered :: Confidence -> Int -> Int -> Double -> Bool
+sufficientlyCovered confidence n k p =
+  -- Accept the coverage if, with high confidence, the actual probability is
+  -- at least 0.9 times the required one.
+  wilsonLow (fromIntegral k) (fromIntegral n) (1 / fromIntegral err) >= tol * p
+  where
+    err = certainty confidence
+    tol = tolerance confidence
+
+insufficientlyCovered :: Maybe Integer -> Int -> Int -> Double -> Bool
+insufficientlyCovered Nothing n k p =
+  fromIntegral k < p * fromIntegral n
+insufficientlyCovered (Just err) n k p =
+  wilsonHigh (fromIntegral k) (fromIntegral n) (1 / fromIntegral err) < p
+
+-- https://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval#Wilson_score_interval
+-- Note:
+-- https://www.ncss.com/wp-content/themes/ncss/pdf/Procedures/PASS/Confidence_Intervals_for_One_Proportion.pdf
+-- suggests we should use a instead of a/2 for a one-sided test. Look
+-- into this.
+wilson :: Integer -> Integer -> Double -> Double
+wilson k n z =
+  (p + z*z/(2*nf) + z*sqrt (p*(1-p)/nf + z*z/(4*nf*nf)))/(1 + z*z/nf)
+  where
+    nf = fromIntegral n
+    p = fromIntegral k / fromIntegral n
+
+wilsonLow :: Integer -> Integer -> Double -> Double
+wilsonLow k n a = wilson k n (invnormcdf (a/2))
+
+wilsonHigh :: Integer -> Integer -> Double -> Double
+wilsonHigh k n a = wilson k n (invnormcdf (1-a/2))
+
+addCoverageCheck :: Confidence -> State -> Property -> Property
+addCoverageCheck confidence st prop
+  | and [ sufficientlyCovered confidence tot n p
+        | (_, _, tot, n, p) <- allCoverage st ] =
+    -- Note: run prop once more so that we get labels for this test case run
+    once prop
+  | or [ insufficientlyCovered (Just (certainty confidence)) tot n p
+       | (_, _, tot, n, p) <- allCoverage st ] =
+    let (labels, tables) = labelsAndTables st in
+    foldr counterexample (property failed{P.reason = "Insufficient coverage"})
+      (paragraphs [labels, tables])
+  | otherwise = prop
+
+allCoverage :: State -> [(Maybe String, String, Int, Int, Double)]
+allCoverage st =
+  [ (key, value, tot, n, p)
+  | ((key, value), p) <- Map.toList (S.requiredCoverage st),
+    let tot =
+          case key of
+            Just key -> Map.findWithDefault 0 key totals
+            Nothing -> numSuccessTests st,
+    let n = Map.findWithDefault 0 value (Map.findWithDefault Map.empty key combinedCounts) ]
+  where
+    combinedCounts :: Map (Maybe String) (Map String Int)
+    combinedCounts =
+      Map.insert Nothing (S.classes st)
+        (Map.mapKeys Just (S.tables st))
+
+    totals :: Map String Int
+    totals = fmap (sum . Map.elems) (S.tables st)
+
 --------------------------------------------------------------------------
 -- the end.
